##############################################################################
#
# Copyright (C) Zenoss, Inc. 2013-2016, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is installed.
#
##############################################################################

import logging
log = logging.getLogger('zen.AWS')

import datetime
import time
import calendar

from cStringIO import StringIO
from lxml import etree

from twisted.internet import defer
from twisted.internet.defer import inlineCallbacks

from zope.component import adapts
from zope.interface import implements

from Products.ZenEvents import ZenEventClasses
from Products.Zuul.form import schema
from Products.Zuul.infos import ProxyProperty
from Products.Zuul.infos.template import RRDDataSourceInfo
from Products.Zuul.interfaces import IRRDDataSourceInfo
from Products.Zuul.utils import ZuulMessageFactory as _t

from ZenPacks.zenoss.PythonCollector.datasources.PythonDataSource \
    import PythonDataSource, PythonDataSourcePlugin

from ZenPacks.zenoss.AWS.utils import awsUrlSign, iso8601, result_errmsg,\
    lookup_cwregion, getPage


class AmazonCloudWatchDataSource(PythonDataSource):
    '''
    Datasource used to capture datapoints from Amazon CloudWatch.
    '''

    ZENPACKID = 'ZenPacks.zenoss.AWS'

    sourcetypes = ('Amazon CloudWatch',)
    sourcetype = sourcetypes[0]

    # RRDDataSource
    component = '${here/id}'
    cycletime = 300

    # PythonDataSource
    plugin_classname = 'ZenPacks.zenoss.AWS.datasources.'\
        'AmazonCloudWatchDataSource.AmazonCloudWatchDataSourcePlugin'

    # AmazonCloudWatchDataSource
    namespace = ''
    metric = ''
    statistic = 'Average'
    dimension = '${here/getDimension}'
    region = '${here/getRegionId}'

    _properties = PythonDataSource._properties + (
        {'id': 'namespace', 'type': 'string'},
        {'id': 'metric', 'type': 'string'},
        {'id': 'statistic', 'type': 'string'},
        {'id': 'dimension', 'type': 'string'},
        {'id': 'region', 'type': 'string'},
        )


class IAmazonCloudWatchDataSourceInfo(IRRDDataSourceInfo):
    '''
    API Info interface for AmazonCloudWatchDataSource.
    '''

    namespace = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Namespace'))

    metric = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Metric Name'))

    statistic = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Statistic'))

    dimension = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Dimension'))

    region = schema.TextLine(
        group=_t('Amazon CloudWatch'),
        title=_t('Region'))


class AmazonCloudWatchDataSourceInfo(RRDDataSourceInfo):
    '''
    API Info adapter factory for AmazonCloudWatchDataSource.
    '''

    implements(IAmazonCloudWatchDataSourceInfo)
    adapts(AmazonCloudWatchDataSource)

    testable = False

    cycletime = ProxyProperty('cycletime')

    namespace = ProxyProperty('namespace')
    metric = ProxyProperty('metric')
    statistic = ProxyProperty('statistic')
    dimension = ProxyProperty('dimension')
    region = ProxyProperty('region')


def timestamp_from_cloudwatch(cloudwatch_time_string):
    try:
        return calendar.timegm(
            time.strptime(cloudwatch_time_string, '%Y-%m-%dT%H:%M:%SZ'))

    except Exception:
        pass

    return calendar.timegm(datetime.datetime.utcnow().timetuple())


class AmazonCloudWatchDataSourcePlugin(PythonDataSourcePlugin):
    proxy_attributes = (
        'ec2accesskey', 'ec2secretkey', 'zAWSCloudWatchSSL',
        'zAWSCloudWatchMaxParallel', 'zAWSCloudWatchMaxRetries'
        )

    @classmethod
    def params(cls, datasource, context):
        return {
            'namespace': datasource.talesEval(datasource.namespace, context),
            'metric': datasource.talesEval(datasource.metric, context),
            'statistic': datasource.talesEval(datasource.statistic, context),
            'dimension': datasource.talesEval(datasource.dimension, context),
            'region': datasource.talesEval(datasource.region, context),
            }

    def ds_description(self, config, ds):
        return '[%s/%s: %s %s/%s for %s]' % (
            config.id,
            ds.params['region'],
            ds.params['statistic'],
            ds.params['namespace'],
            ds.params['metric'],
            ds.params['dimension'] or 'region')

    @inlineCallbacks
    def doQuery(self, config, url, ds):
        try:
            key = (lookup_cwregion(ds.params['region']), ds.ec2accesskey)

            result = yield getPage(
                url,
                concurrent_key=key,
                concurrent_limit=ds.zAWSCloudWatchMaxParallel,
                max_retries=ds.zAWSCloudWatchMaxRetries,
                description=self.ds_description(config, ds))

            defer.returnValue((ds, result))
        except Exception, e:
            defer.returnValue((ds, e))

    @inlineCallbacks
    def collect(self, config):
        ds0 = config.datasources[0]
        log.debug("Collect for AmazonCloudWatch (%d)" % ds0.zAWSCloudWatchMaxParallel)

        results = []

        if not config.datasources[0].component:
            log.warn("Amazon CloudWatch data source type "
                     "not suitable to use on device level")
            defer.returnValue(results)
            return

        # Static for performance collection
        httpVerb = 'GET'
        uriRequest = '/'

        baseRequest = {}
        baseRequest['SignatureMethod'] = 'HmacSHA256'
        baseRequest['SignatureVersion'] = '2'

        deferreds = []

        for ds in config.datasources:
            # CloudWatch only accepts periods that are evenly divisible by 60.
            cycletime = (ds.cycletime / 60) * 60

            hostHeader = lookup_cwregion(ds.params['region'])

            monitorRequest = baseRequest.copy()
            monitorRequest['Action'] = 'GetMetricStatistics'
            monitorRequest['Version'] = '2010-08-01'
            monitorRequest['Period'] = cycletime
            monitorRequest['StartTime'] = iso8601(seconds_ago=(cycletime * 2))
            monitorRequest['EndTime'] = iso8601()
            monitorRequest['Namespace'] = ds.params['namespace']
            monitorRequest['MetricName'] = ds.params['metric']
            monitorRequest['Statistics.member.1'] = ds.params['statistic']

            if ds.params['dimension']:
                dim_group = ds.params['dimension'].split(';')

                i = 0
                for dim_name, dim_value in [x.split('=') for x in dim_group]:
                    i += 1
                    monitorRequest['Dimensions.member.%d.Name' % i] = dim_name
                    monitorRequest['Dimensions.member.%d.Value' % i] = dim_value

            getURL = awsUrlSign(
                httpVerb,
                hostHeader,
                uriRequest,
                monitorRequest,
                (ds.ec2accesskey, ds.ec2secretkey))

            getURL = '{0}://{1}'.format(
                ('https' if ds.zAWSCloudWatchSSL else 'http'), getURL
            )

            d = self.doQuery(config, getURL, ds)
            deferreds.append(d)

        # allow all the requests to complete.
        results = yield defer.DeferredList(deferreds, consumeErrors=True)

        # and return their results
        defer.returnValue([x[1] for x in results])

    def onSuccess(self, results, config):
        data = self.new_data()

        if not config.datasources[0].component:
            data['events'].append({
                'device': config.id,
                'summary': "Amazon CloudWatch data source type "
                           "not suitable to use on device level",
                'severity': ZenEventClasses.Info,
                'eventKey': 'awsCloudWatchCollectionDevice',
                'eventClassKey': 'AWSCloudWatchDevice',
                'eventClass': '/AWS',
                })

            return data

        for ds, result in results:
            if isinstance(result, Exception):
                try:
                    code = getattr(result, 'status', None)
                    error_message = "%s %s: %s" % (code, result.message, result.response)
                except AttributeError:
                    error_message = str(result)

                if 'InvalidInstanceID.NotFound' in error_message:
                    # these are relatively harmless - indicate that a monitored
                    # instance no longer exists.  Ignore and move on.
                    continue

                log.error("Error collecting %s: %s", self.ds_description(config, ds), error_message)
                continue

            if result is None:
                log.info("No response was received for %s", self.ds_description(config, ds))
                continue

            # Only one namespace is used. Mangle the xmlns attribute
            # to make further processing of the XML simpler.
            result = result.replace(' xmlns=', ' xmlnamespace=', 1)

            try:
                stats = etree.parse(StringIO(result))
            except Exception:
                log.debug(
                    '%s (%s): error parsing response XML\n%s',
                    config.id, ds.params['region'], result)
                continue

            if ds == 'volumestatus':
                # Parse volume status events
                try:
                    volumes = stats.xpath('//volumeStatusSet/item')

                    for vol in volumes:
                        volumeID = str(vol.xpath('volumeId[last()]/text()')[0])
                        volumeStatus = str(vol.xpath(
                            'volumeStatus/status/text()'
                        )[0])

                        if volumeStatus == 'ok':
                            data['events'].append({
                                'component': volumeID,
                                'device': config.id,
                                'summary': 'AWS Volume Status: OK',
                                'severity': ZenEventClasses.Clear,
                                'eventClass': '/Status',
                                'eventClassKey': 'AWSVolume',
                            })
                        else:
                            data['events'].append({
                                'component': volumeID,
                                'device': config.id,
                                'summary': "AWS Volume Status: {}".format(
                                    volumeStatus
                                ),
                                'severity': ZenEventClasses.Critical,
                                'eventClass': '/Status',
                                'eventClassKey': 'AWSVolume',
                            })

                except IndexError:
                    continue

            else:
                # Parse cloudwatch metrics
                try:
                    timestamp = timestamp_from_cloudwatch(
                        stats.xpath('//Timestamp[last()]/text()')[0])

                    value = stats.xpath('//%s[last()]/text()' % (
                        ds.params['statistic']))[0]

                except IndexError:
                    # No value in response. This is usually normal.
                    continue

                data['values'][ds.component][ds.datasource] = (value, timestamp)

        data['events'].append({
            'device': config.id,
            'summary': 'AWS CloudWatch: successful metrics collection',
            'severity': ZenEventClasses.Clear,
            'eventKey': 'awsCloudWatchCollection',
            'eventClassKey': 'AWSCloudWatchSuccess',
            })
        return data

    def onError(self, result, config):

        errmsg = 'AWS: %s' % result_errmsg(result)
        data = self.new_data()

        if 'timeout' in errmsg or 'connection lost' in errmsg:
            log.debug('%s: %s', config.id, errmsg)
        else:
            log.error('%s: %s', config.id, errmsg)
            data['events'].append({
                'device': config.id,
                'summary': errmsg,
                'severity': ZenEventClasses.Error,
                'eventKey': 'awsCloudWatchCollection',
                'eventClassKey': 'AWSCloudWatchError',
                })

        return data
