##############################################################################
#
# Copyright (C) Zenoss, Inc. 2013-2015, all rights reserved.
#
# This content is made available according to terms specified in
# License.zenoss under the directory where your Zenoss product is installed.
#
##############################################################################

'''
Collects information about AWS S3 buckets.
'''

import collections
import json

from itertools import chain
from logging import getLogger
log = getLogger('zen.ZenModeler')

from socket import gaierror
from twisted.internet import defer

from Products.DataCollector.plugins.CollectorPlugin import PythonPlugin
from Products.DataCollector.plugins.DataMaps import ObjectMap, RelationshipMap

from Products.ZenUtils.Utils import prepId

from ZenPacks.zenoss.AWS import MODULE_NAME
from ZenPacks.zenoss.AWS.utils import addLocalLibPath, format_time

addLocalLibPath()

from boto.ec2.connection import EC2ResponseError
from boto.s3.connection import S3Connection
import boto

from ZenPacks.zenoss.AWS.txboto import TwistedS3Connection

'''
Models S3 Buckets for an Amazon EC2 account.
'''


class S3Buckets(PythonPlugin):
    deviceProperties = PythonPlugin.deviceProperties + (
        'ec2accesskey',
        'ec2secretkey',
    )

    @defer.inlineCallbacks
    def collect(self, device, log):
        log.info(
            'Modeler %s collecting data for device %s',
            self.name(), device.id)

        accesskey = getattr(device, 'ec2accesskey', None)
        secretkey = getattr(device, 'ec2secretkey', None)

        old_logger = boto.log.error
        boto.log.error = lambda x: x
        try:
            if not secretkey or not accesskey:
                raise EC2ResponseError('', '', '')
            s3connection = TwistedS3Connection.adapt(
                S3Connection(accesskey, secretkey))
        except EC2ResponseError:
            log.error('Invalid Keys. '
                      'Check your EC2 Access Key and EC2 Secret Key.')
            return
        except gaierror, e:
            log.error(e.strerror)
            return
        finally:
            boto.log.error = old_logger

        # S3Buckets
        buckets = yield s3connection.get_all_buckets()

        # Unfortunately, determining what region each bucket is in
        # requires an additional API call.   These are done in parallel
        # to achieve reasonable performance.
        bucket_region = {}
        deferreds = []
        names = []
        for bucket in buckets:
            d = bucket.get_location()
            names.append(bucket.name)
            deferreds.append(d)

        results = yield defer.DeferredList(deferreds, consumeErrors=True)

        for i, value in enumerate(results):
            success, result = value
            name = names[i]

            if success:
                # value is region.
                if result == "":
                    # this is the default "US Standard" region for S3 buckets
                    # that do not have a LocationConstraint.
                    result = "us-east-1"

                bucket_region.update({name: result})
            else:
                # result is a twisted Failure object, and result.value is the
                # exception that occurred.
                log.warning("Unable to determine region of S3 bucket %s: %s", name, str(result.value))

        results = {
            'all_buckets': buckets,
            'bucket_region': bucket_region
        }
        defer.returnValue(results)

    def process(self, device, results, log):
        log.info(
            'Modeler %s processing data for device %s',
            self.name(), device.id)

        maps = collections.OrderedDict([
            ('s3buckets', []),
        ])

        maps['s3buckets'].append(
            s3buckets_rm(
                results['all_buckets'],
                results['bucket_region']
            )
        )

        return list(chain.from_iterable(maps.itervalues()))


def s3buckets_rm(buckets, bucket_region):
    '''
    Return buckets RelationshipMap given a BucketInfo
    ResultSet and a mapping of bucket id to region
    '''
    bucket_oms = []
    for bucket in buckets:
        bucket_oms.append(ObjectMap(data={
            'id': prepId(bucket.name),
            'title': bucket.name,
            'region': bucket_region.get(bucket.name, ""),
            'creation_date': format_time(bucket.creation_date),
        }))

    return RelationshipMap(
        relname='s3buckets',
        modname=MODULE_NAME['S3Bucket'],
        objmaps=bucket_oms
    )
